/////////1/////////2/////////3/////////4/////////5/////////6/////////7/////////8
// portable_binary_oarchive.cpp

// (C) Copyright 2002-7 Robert Ramey - http://www.rrsd.com .
// Copyright (c) 2007-2015 Hartmut Kaiser
// Use, modification and distribution is subject to the Boost Software
// License, Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)

//  See http://www.boost.org for updates, documentation, and revision history.

#include <boost/version.hpp>
#include <hpx/config.hpp>

// export the defined functions
#define BOOST_ARCHIVE_SOURCE

// this hack is needed to properly compile this shared library, allowing to
// export the symbols and auto link with the serialization
#if !defined(BOOST_ALL_NO_LIB) && !defined(BOOST_SERIALIZATION_NO_LIB)
// Set the name of our library, this will get undef'ed by auto_link.hpp
// once it's done with it:
#define BOOST_LIB_NAME boost_serialization

// If we're importing code from a dll, then tell auto_link.hpp about it:
#if defined(BOOST_ALL_DYN_LINK) || defined(BOOST_SERIALIZATION_DYN_LINK)
#  define BOOST_DYN_LINK
#endif

// And include the header that does the work:
#include <boost/config/auto_link.hpp>
#endif  // auto-linking disabled

#include <ostream>

#include <boost/detail/endian.hpp>
#include <hpx/util/portable_binary_oarchive.hpp>

#if defined(BOOST_MSVC)
#  include <intrin.h>
#  pragma intrinsic(memcpy)
#  pragma intrinsic(memset)
#endif

namespace hpx { namespace util
{

void portable_binary_oarchive::save_impl(
    boost::int64_t const l, char const maxsize)
{
#if defined(HPX_DEBUG_SERIALIZATION)
    static char const type = 'i';
    this->primitive_base_t::save(type);
    this->primitive_base_t::save(maxsize);
#endif

    char size = 0;
    if (l == 0) {
        this->primitive_base_t::save(size);
        return;
    }

    boost::int64_t ll;
    bool negative = (l < 0);
    if (negative)
        ll = -l;
    else
        ll = l;

    do {
        ll >>= CHAR_BIT;
        ++size;
    } while (ll != 0);

    HPX_ASSERT(size <= maxsize);

    this->primitive_base_t::save(size);
    this->primitive_base_t::save(negative);

    if(negative)
        ll = -l;
    else
        ll = l;

    char* cptr = reinterpret_cast<char *>(& ll); //-V206
#ifdef BOOST_BIG_ENDIAN
    cptr += (sizeof(boost::int64_t) - size);
    if(this->flags() & endian_little)
        reverse_bytes(size, cptr);
#else
    if(this->flags() & endian_big)
        reverse_bytes(size, cptr);
#endif

    this->primitive_base_t::save_binary(cptr, static_cast<std::size_t>(size));
}

void portable_binary_oarchive::save_impl(
    boost::uint64_t const ul, char const maxsize)
{
#if defined(HPX_DEBUG_SERIALIZATION)
    static char const type = 'u';
    this->primitive_base_t::save(type);
    this->primitive_base_t::save(maxsize);
#endif

    char size = 0;
    if (ul == 0) {
        this->primitive_base_t::save(size);
        return;
    }

    boost::uint64_t ll = ul;
    do {
        ll >>= CHAR_BIT;
        ++size;
    } while (ll != 0);

    HPX_ASSERT(size <= maxsize);
    this->primitive_base_t::save(size);

    ll = ul;
    char* cptr = reinterpret_cast<char*>(&ll); //-V206

#ifdef BOOST_BIG_ENDIAN
    cptr += (sizeof(boost::uint64_t) - size);
    if(this->flags() & endian_little)
        reverse_bytes(size, cptr);
#else
    if(this->flags() & endian_big)
        reverse_bytes(size, cptr);
#endif

    this->primitive_base_t::save_binary(cptr, static_cast<std::size_t>(size));
}

#if defined(__GNUG__) && !defined(__INTEL_COMPILER)
#   if defined(HPX_GCC_DIAGNOSTIC_PRAGMA_CONTEXTS)
#       pragma GCC diagnostic push
#   endif
#   pragma GCC diagnostic ignored "-Wconversion"
#endif
void portable_binary_oarchive::init(util::binary_filter* filter, unsigned int flags)
{
    if ((this->flags() & (endian_big | endian_little)) == (endian_big | endian_little))
    {
        // bail out if both flags are specified
        BOOST_THROW_EXCEPTION(
            portable_binary_oarchive_exception());
    }

    if (!(flags & boost::archive::no_header)) {
        // write signature in an archive version independent manner
        const std::string file_signature(
            boost::archive::BOOST_ARCHIVE_SIGNATURE());
        *this << file_signature;

        // write library version
        const boost::archive::version_type v(
            boost::archive::BOOST_ARCHIVE_VERSION());
        *this << v;
    }

    save(static_cast<boost::uint16_t>(this->flags() >> CHAR_BIT));

    // handle filter and compression in the archive separately
    bool has_filter = filter != 0;
    save(has_filter);

    if (has_filter && (this->flags() & enable_compression)) {
        *this << filter;
        this->set_filter(filter);
    }
}
#if defined(__GNUG__) && !defined(__INTEL_COMPILER)
#   if defined(HPX_GCC_DIAGNOSTIC_PRAGMA_CONTEXTS)
#       pragma GCC diagnostic pop
#   endif
#endif

}}

#include <boost/archive/detail/archive_serializer_map.hpp>
#include <boost/archive/impl/archive_serializer_map.ipp>

namespace boost {
namespace archive {

    template class HPX_ALWAYS_EXPORT
        detail::archive_serializer_map<hpx::util::portable_binary_oarchive>;

} // namespace archive
} // namespace boost

#include <hpx/util/basic_binary_oprimitive_impl.hpp>

namespace hpx { namespace util
{
    // explicitly instantiate base primitive
    template class basic_binary_oprimitive<
        hpx::util::portable_binary_oarchive
    >;
}}
